@@ -64,6 +64,7 @@ typedef struct token *attr_t(struct token *, struct symbol *,
 static attr_t
 	attribute_packed, attribute_aligned, attribute_modifier,
 	attribute_address_space, attribute_context,
+	attribute_designated_init,
 	attribute_transparent_union, ignore_attribute,
 	attribute_mode, attribute_force;
 
@@ -319,6 +320,10 @@ static struct symbol_op context_op = {
 	.attribute = attribute_context,
 };
 
+static struct symbol_op designated_init_op = {
+	.attribute = attribute_designated_init,
+};
+
 static struct symbol_op transparent_union_op = {
 	.attribute = attribute_transparent_union,
 };
@@ -391,6 +396,7 @@ static struct init_keyword {
 
 	/* Predeclared types */
 	{ "__builtin_va_list", NS_TYPEDEF, .type = &ptr_ctype, .op = &spec_op },
+	{ "__builtin_ms_va_list", NS_TYPEDEF, .type = &ptr_ctype, .op = &spec_op },
 	{ "__int128_t",	NS_TYPEDEF, .type = &lllong_ctype, .op = &spec_op },
 	{ "__uint128_t",NS_TYPEDEF, .type = &ulllong_ctype, .op = &spec_op },
 
@@ -453,7 +459,10 @@ static struct init_keyword {
 	{ "address_space",NS_KEYWORD,	.op = &address_space_op },
 	{ "mode",	NS_KEYWORD,	.op = &mode_op },
 	{ "context",	NS_KEYWORD,	.op = &context_op },
+	{ "designated_init",	NS_KEYWORD,	.op = &designated_init_op },
 	{ "__transparent_union__",	NS_KEYWORD,	.op = &transparent_union_op },
+	{ "noreturn",	NS_KEYWORD,	MOD_NORETURN,	.op = &attr_mod_op },
+	{ "__noreturn__",	NS_KEYWORD,	MOD_NORETURN,	.op = &attr_mod_op },
 
 	{ "__mode__",	NS_KEYWORD,	.op = &mode_op },
 	{ "QI",		NS_KEYWORD,	MOD_CHAR,	.op = &mode_QI_op },
@@ -468,85 +477,107 @@ static struct init_keyword {
 	{ "__TI__",	NS_KEYWORD,	MOD_LONGLONGLONG,	.op = &mode_TI_op },
 	{ "word",	NS_KEYWORD,	MOD_LONG,	.op = &mode_word_op },
 	{ "__word__",	NS_KEYWORD,	MOD_LONG,	.op = &mode_word_op },
-
-	/* Ignored attributes */
-	{ "nothrow",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__nothrow",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__nothrow__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "malloc",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__malloc__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "nonnull",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__nonnull",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__nonnull__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "format",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__format__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "format_arg",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__format_arg__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "section",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__section__",NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "unused",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__unused__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "const",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__const",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__const__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "noreturn",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__noreturn__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "no_instrument_function",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__no_instrument_function__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "sentinel",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__sentinel__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "regparm",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__regparm__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "weak",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__weak__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "alias",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__alias__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "pure",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__pure__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "always_inline",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__always_inline__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "syscall_linkage",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__syscall_linkage__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "visibility",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__visibility__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "deprecated",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__deprecated__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "noinline",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__noinline__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "used",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__used__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "warn_unused_result",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__warn_unused_result__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "model",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__model__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "cdecl",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__cdecl__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "stdcall",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__stdcall__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "fastcall",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__fastcall__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "dllimport",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__dllimport__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "dllexport",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__dllexport__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "constructor",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__constructor__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "destructor",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__destructor__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "cold",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__cold__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "hot",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__hot__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "warning",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__warning__",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "bounded",	NS_KEYWORD,	.op = &ignore_attr_op },
-	{ "__bounded__",	NS_KEYWORD,	.op = &ignore_attr_op },
 };
 
+const char *ignored_attributes[] = {
+	"alias",
+	"__alias__",
+	"alloc_size",
+	"__alloc_size__",
+	"always_inline",
+	"__always_inline__",
+	"bounded",
+	"__bounded__",
+	"cdecl",
+	"__cdecl__",
+	"cold",
+	"__cold__",
+	"const",
+	"__const",
+	"__const__",
+	"constructor",
+	"__constructor__",
+	"deprecated",
+	"__deprecated__",
+	"destructor",
+	"__destructor__",
+	"dllexport",
+	"__dllexport__",
+	"dllimport",
+	"__dllimport__",
+	"externally_visible",
+	"__externally_visible__",
+	"fastcall",
+	"__fastcall__",
+	"format",
+	"__format__",
+	"format_arg",
+	"__format_arg__",
+	"hot",
+	"__hot__",
+	"l1_text",
+	"__l1_text__",
+	"l1_data",
+	"__l1_data__",
+	"l2",
+	"__l2__",
+	"may_alias",
+	"__may_alias__",
+	"malloc",
+	"__malloc__",
+	"may_alias",
+	"__may_alias__",
+	"model",
+	"__model__",
+	"ms_abi",
+	"__ms_abi__",
+	"naked",
+	"__naked__",
+	"no_instrument_function",
+	"__no_instrument_function__",
+	"noinline",
+	"__noinline__",
+	"nonnull",
+	"__nonnull",
+	"__nonnull__",
+	"nothrow",
+	"__nothrow",
+	"__nothrow__",
+	"pure",
+	"__pure__",
+	"regparm",
+	"__regparm__",
+	"section",
+	"__section__",
+	"sentinel",
+	"__sentinel__",
+	"signal",
+	"__signal__",
+	"stdcall",
+	"__stdcall__",
+	"syscall_linkage",
+	"__syscall_linkage__",
+	"sysv_abi",
+	"__sysv_abi__",
+	"unused",
+	"__unused__",
+	"used",
+	"__used__",
+	"visibility",
+	"__visibility__",
+	"warn_unused_result",
+	"__warn_unused_result__",
+	"warning",
+	"__warning__",
+	"weak",
+	"__weak__",
+};
+
+
 void init_parser(int stream)
 {
 	int i;
-	for (i = 0; i < sizeof keyword_table/sizeof keyword_table[0]; i++) {
+	for (i = 0; i < ARRAY_SIZE(keyword_table); i++) {
 		struct init_keyword *ptr = keyword_table + i;
 		struct symbol *sym = create_symbol(stream, ptr->name, SYM_KEYWORD, ptr->ns);
 		sym->ident->keyword = 1;
@@ -556,8 +587,17 @@ void init_parser(int stream)
 		sym->ctype.base_type = ptr->type;
 		sym->op = ptr->op;
 	}
+
+	for (i = 0; i < ARRAY_SIZE(ignored_attributes); i++) {
+		const char * name = ignored_attributes[i];
+		struct symbol *sym = create_symbol(stream, name, SYM_KEYWORD,
+						   NS_KEYWORD);
+		sym->ident->keyword = 1;
+		sym->op = &ignore_attr_op;
+	}
 }
 
+
 // Add a symbol to the list of function-local symbols
 static void fn_local_symbol(struct symbol *sym)
 {
@@ -1127,6 +1167,15 @@ static struct token *attribute_context(struct token *token, struct symbol *attr,
 	return token;
 }
 
+static struct token *attribute_designated_init(struct token *token, struct symbol *attr, struct decl_state *ctx)
+{
+	if (ctx->ctype.base_type && ctx->ctype.base_type->type == SYM_STRUCT)
+		ctx->ctype.base_type->designated_init = 1;
+	else
+		warning(token->pos, "attribute designated_init applied to non-structure type");
+	return token;
+}
+
 static struct token *attribute_transparent_union(struct token *token, struct symbol *attr, struct decl_state *ctx)
 {
 	if (Wtransparent_union)
@@ -1659,6 +1708,7 @@ static struct token *pointer(struct token *token, struct decl_state *ctx)
 		ctx->ctype.base_type = ptr;
 		ctx->ctype.as = 0;
 		ctx->ctype.contexts = NULL;
+		ctx->ctype.alignment = 0;
 
 		token = handle_qualifiers(token->next, ctx);
 		ctx->ctype.base_type->endpos = token->pos;
@@ -1849,13 +1899,35 @@ static struct token *parse_asm_clobbers(struct token *token, struct statement *s
 	return token;
 }
 
+static struct token *parse_asm_labels(struct token *token, struct statement *stmt,
+		        struct symbol_list **labels)
+{
+	struct symbol *label;
+
+	do {
+		token = token->next; /* skip ':' and ',' */
+		if (token_type(token) != TOKEN_IDENT)
+			return token;
+		label = label_symbol(token);
+		add_symbol(labels, label);
+		token = token->next;
+	} while (match_op(token, ','));
+	return token;
+}
+
 static struct token *parse_asm_statement(struct token *token, struct statement *stmt)
 {
+	int is_goto = 0;
+
 	token = token->next;
 	stmt->type = STMT_ASM;
 	if (match_idents(token, &__volatile___ident, &__volatile_ident, &volatile_ident, NULL)) {
 		token = token->next;
 	}
+	if (token_type(token) == TOKEN_IDENT && token->ident == &goto_ident) {
+		is_goto = 1;
+		token = token->next;
+	}
 	token = expect(token, '(', "after asm");
 	token = parse_expression(token, &stmt->asm_string);
 	if (match_op(token, ':'))
@@ -1864,6 +1936,8 @@ static struct token *parse_asm_statement(struct token *token, struct statement *
 		token = parse_asm_operands(token, stmt, &stmt->asm_inputs);
 	if (match_op(token, ':'))
 		token = parse_asm_clobbers(token, stmt, &stmt->asm_clobbers);
+	if (is_goto && match_op(token, ':'))
+		token = parse_asm_labels(token, stmt, &stmt->asm_labels);
 	token = expect(token, ')', "after asm");
 	return expect(token, ';', "at end of asm-statement");
 }
